@@ -16,14 +16,15 @@
 	if (rc) {
 		int fd = fileno(rc);
 
-		pseudo_debug(2, "freopen64 '%s': fd %d\n", path, fd);
+		pseudo_debug(PDBGF_FILE, "freopen64 '%s': fd %d\n", path, fd);
 		if (real___fxstat64(_STAT_VER, fd, &buf) != -1) {
 			if (!existed) {
+				real_fchmod(fd, PSEUDO_FS_MODE(0600, 0));
 				pseudo_client_op(OP_CREAT, 0, -1, -1, path, &buf);
 			}
 			pseudo_client_op(OP_OPEN, pseudo_access_fopen(mode), fd, -1, path, &buf);
 		} else {
-			pseudo_debug(1, "fopen (fd %d) succeeded, but stat failed (%s).\n",
+			pseudo_debug(PDBGF_FILE, "fopen (fd %d) succeeded, but stat failed (%s).\n",
 				fd, strerror(errno));
 			pseudo_client_op(OP_OPEN, pseudo_access_fopen(mode), fd, -1, path, 0);
 		}
